#include <stdio.h>      // fprintf()
#include <stdlib.h>     // malloc()
#include <string.h>     // memcpy()
#include <stdint.h>     // uintptr_t
#include <assert.h>
#include "gc.h"


// tunables
#define ARENASIZE 1048576       // size of each allocation arena
#define GCPERIOD 1              // call gc() every GCPERIOD grow()s

// these should not be touched
#define MAXSIZE 65535           // maximum size of an allocation object
#define ALIGN 4                 // align all objects on ALIGN-byte boundaries
#define MAGIC 0x55              // 7-bit magic number to identify heap objects


// bespoke garbage collector for new(); idea:
//   * allocate medium-sized blocks (arenas) of mem in a linked list
//   * new() always takes from the end of the latest block
//   * gc() scans active set from stack and registers (like boehm-gc)
//   * arenas are freed when they become vacant (home to no active objects)
//   * externally-managed areas can be hooked into scan/gc.

typedef struct object object_t;
typedef struct arena arena_t;
struct arena {
    arena_t *next;
    size_t nobjects;
    size_t nbytes;
    int active:1;               // used during GC
    object_t *p;                // current position in mem (first object)
    uint8_t mem[];              // filled from the end with objects (see below)
};

// memory layout:
// each arena starts with the metadata as defined in the struct above.
// the 'mem' area consists of a sequence of allocation objects of the following
// format.
//
//   object ::= 0x00 magic mark size->L byte{L}
//   magic  ::= 0x55            -- 7 bits
//   mark   ::= 0|1             -- 1 bit, used during GC
//   size   ::= 0x0001-0xFFFF   -- 16 bits
//
//   arena  ::= object* end
//   end    ::= 0x00 0x55 0 0x0000
//
struct object {
    uint8_t null;               // always 0x00
    uint8_t magic:7;            // always MAGIC
    uint8_t mark:1;             // we don't care whether this is LSB or MSB
    uint16_t size;              // [bytes] size of body (excluding header)
    uint8_t body[];
};

arena_t *arena = NULL;
static void *stackbase = NULL;  // highest stack address
static void *addrmin = NULL;    // lowest possible heap address
static void *addrmax = NULL;    // highest possible heap address + 1

external_t *external = NULL;

// statistics
size_t allocarenas = 0;
size_t allocobjects = 0;
size_t allocbytes = 0;
size_t freedarenas = 0;
size_t freedobjects = 0;
size_t freedbytes = 0;
size_t activearenas = 0;
size_t activeobjects = 0;
size_t activebytes = 0;
size_t peakaarenas = 0;         // active
size_t peakaobjects = 0;
size_t peakabytes = 0;
size_t peakrarenas = 0;         // resident
size_t peakrobjects = 0;
size_t peakrbytes = 0;

// gc helper: does o have the expected object structure?
static int isobject(object_t *o)
{
    return (o->null == 0 && o->magic == MAGIC);
}

static void scan(void *p, size_t n, uint8_t sense);
static void scanref(void *x, uint8_t sense)
{
    object_t *o;
    arena_t *a;
    external_t *e;

    // handle externals
    for(e=external; e; e=e->next) {
        if(x < e->addrmin || x >= e->addrmax)
            continue;   // not in this area
        if(e->scanref(scan, scanref, e, x, sense))
            return;     // found
    }

    // check that the address x refers to a heap object...

    // check for expected alignment
    if((uintptr_t)x % ALIGN != 0)
        return;

    // check for expected heap address range
    if(x < addrmin || x >= addrmax)
        return;

    // the (presumed) object hosting x
    o = x - sizeof(object_t);

    // find the arena containing o
    for(a=arena; a; a=a->next) {
        if(o >= a->p && (void *)o < (void *)a->mem + ARENASIZE)
            break;
    }
    if(!a)
        return;

    // check for expected object structure on o
    if(!isobject(o))
        return;

    // size should reveal the following object
    if(!isobject(x + o->size))
        return;

    assert(o->body == x);

    // have valid heap object...

    // skip object if already visited
    if(o->mark == sense)
        return;
    else
        o->mark = sense;

    // mark arena active
    a->active = sense;

    if(sense) {
        activebytes += o->size;
        activeobjects += 1;
    }

    // recursively scan the object
    scan(o->body, o->size, sense);
}

// sense=1: mark the active set (assign 1 to mark flags)
// sense=0: clear mark flags (assign 0)
static void scan(void *p, size_t n, uint8_t sense)
{
    // scan the given memory area for pointers into arenas and mark
    n -= sizeof(object_t *) - 1;
    for(; n>0; p++,n--) {
        void *x = *(void **)p;  // the address at p
        scanref(x, sense);      // scan the object, if any, at x
    }
}

// try to guess arch from platform #defines
#if defined __i386__ || defined _WIN32
    #define ARCH_X86
#elif defined __x86_64__ || defined _WIN64
    #define ARCH_AMD64
#endif

static void scanregs(uint8_t sense)
{
    #define SCANREG(REG) scanref(r ## REG, sense);

#if defined ARCH_X86
    #define SAVEREG(REG) \
            void *r ## REG; \
            __asm__("movl %%" #REG ", %[reg]" : [reg] "=m" (r ## REG));
    SAVEREG(eax);
    SAVEREG(ebx);
    SAVEREG(ecx);
    SAVEREG(edx);
    SAVEREG(esi);
    SAVEREG(edi);
    SAVEREG(ebp);
    #undef SAVEREG

    SCANREG(eax);
    SCANREG(ebx);
    SCANREG(ecx);
    SCANREG(edx);
    SCANREG(esi);
    SCANREG(edi);
    SCANREG(ebp);
#elif defined ARCH_AMD64
    #define SAVEREG(REG) \
            void *r ## REG; \
            __asm__("movq %%" #REG ", %[reg]" : [reg] "=m" (r ## REG));
    SAVEREG(rax);
    SAVEREG(rbx);
    SAVEREG(rcx);
    SAVEREG(rdx);
    SAVEREG(rsi);
    SAVEREG(rdi);
    SAVEREG(rbp);
    SAVEREG(r8);
    SAVEREG(r9);
    SAVEREG(r10);
    SAVEREG(r11);
    SAVEREG(r12);
    SAVEREG(r13);
    SAVEREG(r14);
    SAVEREG(r15);
    #undef SAVEREG

    SCANREG(rax);
    SCANREG(rbx);
    SCANREG(rcx);
    SCANREG(rdx);
    SCANREG(rsi);
    SCANREG(rdi);
    SCANREG(rbp);
    SCANREG(r8);
    SCANREG(r9);
    SCANREG(r10);
    SCANREG(r11);
    SCANREG(r12);
    SCANREG(r13);
    SCANREG(r14);
    SCANREG(r15);
#else
    #error "need to #define one of ARCH_{X86,AMD64}"
#endif

    #undef SCANREG
}

void gc(void)
{
    if(!arena)
        return;

    void *stacktop = alloca(0);
    assert(stacktop < stackbase);   // grows down

    activebytes = 0;    // filled during scan
    activeobjects = 0;  // filled during scan
    activearenas = 0;

    // scan active set, marking active arenas
    scanregs(1);
    scan(stacktop, stackbase-stacktop, 1);
    // XXX scan global variables

    // free vacant arenas
    arena_t **pa, *a;
    for(pa=&arena; (a=*pa); ) {
        if(a->active || !a->next) { // don't free the last arena
            pa = &a->next;          // skip this arena

            activearenas += 1;
        } else {
            *pa = a->next;          // remove this arena from the list

            freedbytes += a->nbytes;
            freedobjects += a->nobjects;
            freedarenas += 1;

            free(a);
        }
    }

    if(activebytes > peakabytes) {
        peakabytes = activebytes;
        peakaobjects = activeobjects;
        peakaarenas = activearenas;
    }

    // sweep externals
    external_t *e;
    for(e=external; e; e=e->next)
        e->sweep(e);

    // clear mark bits
    scanregs(0);
    scan(stacktop, stackbase-stacktop, 0);
}

void grow(void)
{
    if(allocarenas % GCPERIOD == 0)
        gc();

    arena_t *a = malloc(sizeof(arena_t) + ARENASIZE + sizeof(object_t));
    assert(a);

    a->next = arena;
    a->nobjects = 0;
    a->nbytes = 0;
    a->active = 0;
    a->p = (void *)a->mem + ARENASIZE;

    // fill in the 'end' object
    a->p->null = 0;
    a->p->magic = MAGIC;
    a->p->mark = 0;
    a->p->size = 0;

    arena = a;

    // XXX ensure alignment of arena->p
    assert(sizeof(object_t) == 4);
    assert(sizeof(object_t) % ALIGN == 0);
    assert(ARENASIZE % ALIGN == 0);
    assert((uintptr_t)arena->p % ALIGN == 0);

    // update addrmin/addrmax
    if(!addrmin || (void *)a->mem < addrmin)
        addrmin = a->mem;
    if((void *)a + ARENASIZE > addrmax)
        addrmax = (void *)a->mem + ARENASIZE;

    allocarenas += 1;
}

void *new(size_t size)
{
    assert(arena);
    assert(size <= MAXSIZE);
    assert(size > 0);

    // grow as needed
    size_t osize = sizeof(object_t) + size;
    if(osize > (uintptr_t)arena->p - (uintptr_t)arena->mem)
        grow();
    assert(osize <= (uintptr_t)arena->p - (uintptr_t)arena->mem);

    // ensure alignment
    assert((uintptr_t)arena->p % ALIGN == 0);
    assert(osize % ALIGN == 0);

    object_t *obj = (void *)arena->p - osize;
    assert((void *)obj >= (void *)arena->mem);
    obj->null = 0;
    obj->magic = MAGIC;
    obj->mark = 0;
    obj->size = size;

    arena->p = obj;
    arena->nobjects += 1;
    arena->nbytes += size;

    allocbytes += size;
    allocobjects += 1;
    if(allocbytes-freedbytes > peakrbytes) {
        peakrbytes = allocbytes-freedbytes;
        peakrobjects = allocobjects-freedobjects;
        peakrarenas = allocarenas-freedarenas;
    }

    return obj->body;
}

void *dup(void *p, size_t size)
{
    void *r = new(size);
    if(p)
        memcpy(r, p, size);
    return r;
}

void gc_add_external(external_t *ext)
{
    ext->next = external;
    external = ext;
}

void gc_init(void *sb)
{
    stackbase = sb;
    grow(); // allocate first arena
}

// print stats
void gc_stats(FILE *f)
{
    size_t peakrwaste = peakrbytes - peakabytes;
    size_t peakroverhead = peakrarenas * ARENASIZE - peakrbytes;
    fprintf(f, "(alloc'd:   %zu bytes in %zu objects over %zu arenas;\n"
               " freed:     %zu bytes in %zu objects over %zu arenas;\n"
               " peak act.: %zu bytes in %zu objects over %zu arenas;\n"
               " peak res.: %zu bytes in %zu objects over %zu arenas,\n"
               "            %.1f%% waste, %.2f%% overhead)\n",
            allocbytes, allocobjects, allocarenas,
            freedbytes, freedobjects, freedarenas,
            peakabytes, peakaobjects, peakaarenas,
            peakrbytes, peakrobjects, peakrarenas,
            peakrwaste*1000/peakrbytes/10.0,
            peakroverhead*10000/peakrbytes/100.0);

    external_t *e;
    for(e=external; e; e=e->next) if(e->stats) e->stats(e, f);
}
